@@ -22,10 +22,10 @@ import asyncio
 import inspect
 import logging
 import ssl
+import random
 from datetime import datetime, timedelta, timezone
 from functools import partial
 from http import HTTPStatus
-from random import randint
 
 import aiohttp
 from lxml.etree import XMLSyntaxError
@@ -34,7 +34,7 @@ from apscheduler.schedulers.asyncio import AsyncIOScheduler
 from openleadr import enums, objects, errors
 from openleadr.messaging import create_message, parse_message, \
                                 validate_xml_schema, validate_xml_signature
-from openleadr.utils import generate_id, certificate_fingerprint, find_by, cron_config
+from openleadr import utils
 
 logger = logging.getLogger('openleadr')
 
@@ -45,7 +45,8 @@ class OpenADRClient:
     you can always choose to call them manually.
     """
     def __init__(self, ven_name, vtn_url, debug=False, cert=None, key=None,
-                 passphrase=None, vtn_fingerprint=None, show_fingerprint=True, ca_file=None):
+                 passphrase=None, vtn_fingerprint=None, show_fingerprint=True, ca_file=None,
+                 allow_jitter=True):
         """
         Initializes a new OpenADR Client (Virtual End Node)
 
@@ -82,11 +83,13 @@ class OpenADRClient:
         self.scheduler = AsyncIOScheduler()
         self.client_session = None
         self.report_queue_task = None
+        self.responded_events = {}              # Holds the events that we already saw.
 
         self.cert_path = cert
         self.key_path = key
         self.passphrase = passphrase
         self.ca_file = ca_file
+        self.allow_jitter = allow_jitter
 
         if cert and key:
             with open(cert, 'rb') as file:
@@ -97,7 +100,7 @@ class OpenADRClient:
                 print("")
                 print("*" * 80)
                 print("Your VEN Certificate Fingerprint is ".center(80))
-                print(f"{certificate_fingerprint(cert).center(80)}".center(80))
+                print(f"{utils.certificate_fingerprint(cert).center(80)}".center(80))
                 print("Please deliver this fingerprint to the VTN.".center(80))
                 print("You do not need to keep this a secret.".center(80))
                 print("*" * 80)
@@ -114,7 +117,7 @@ class OpenADRClient:
         """
         # if not hasattr(self, 'on_event'):
         #     raise NotImplementedError("You must implement on_event.")
-
+        self.loop = asyncio.get_event_loop()
         await self.create_party_registration()
 
         if not self.ven_id:
@@ -124,38 +127,20 @@ class OpenADRClient:
 
         if self.reports:
             await self.register_reports(self.reports)
-            loop = asyncio.get_event_loop()
-            self.report_queue_task = loop.create_task(self._report_queue_worker())
+            self.report_queue_task = self.loop.create_task(self._report_queue_worker())
 
         await self._poll()
 
         # Set up automatic polling
-        if self.poll_frequency.total_seconds() < 60:
-            seconds_offset = randint(0, self.poll_frequency.seconds)
-            cron_second = f"{seconds_offset}/{self.poll_frequency.seconds}"
-            cron_minute = "*"
-            cron_hour = "*"
-        elif self.poll_frequency.total_seconds() < 3600:
-            cron_second = randint(0, 59)
-            cron_minute = f'*/{int(self.poll_frequency.total_seconds() / 60)}'
-            cron_hour = "*"
-        elif self.poll_frequency.total_seconds() < 86400:
-            cron_second = randint(0, 59)
-            cron_minute = "0"
-            cron_hour = f'*/{int(self.poll_frequency.total_seconds() / 3600)}'
-        elif self.poll_frequency.total_seconds() > 86400:
+        if self.poll_frequency > timedelta(hours=24):
             logger.warning("Polling with intervals of more than 24 hours is not supported. "
                            "Will use 24 hours as the logging interval.")
-            cron_second = randint(0, 59)
-            cron_minute = "0"
-            cron_hour = "0"
-            return
+            self.poll_frequency = timedelta(hours=24)
+        cron_config = utils.cron_config(self.poll_frequency, randomize_seconds=self.allow_jitter)
 
         self.scheduler.add_job(self._poll,
                                trigger='cron',
-                               second=cron_second,
-                               minute=cron_minute,
-                               hour=cron_hour)
+                               **cron_config)
         self.scheduler.start()
 
     async def stop(self):
@@ -258,38 +243,46 @@ class OpenADRClient:
             item_base = None
         elif isinstance(measurement, objects.Measurement):
             item_base = measurement
+        elif isinstance(measurement, dict):
+            utils.validate_report_measurement_dict(measurement)
+            power_attributes = object.PowerAttributes(**measurement.get('power_attributes')) or None
+            item_base = objects.Measurement(name=measurement['name'],
+                                            description=measurement['description'],
+                                            unit=measurement['unit'],
+                                            scale=measurement.get('scale'),
+                                            power_attributes=power_attributes)
         elif measurement.upper() in enums.MEASUREMENTS.members:
             item_base = enums.MEASUREMENTS[measurement.upper()]
         else:
-            item_base = objects.Measurement(item_name='customUnit',
-                                            item_description=measurement,
-                                            item_units=unit,
-                                            si_scale_code=scale)
+            item_base = objects.Measurement(name='customUnit',
+                                            description=measurement,
+                                            unit=unit,
+                                            scale=scale)
 
         if report_name != 'TELEMETRY_STATUS' and scale is not None:
-            if scale in enums.SI_SCALE_CODE.values:
-                item_base.si_scale_code = scale
+            if item_base.scale is not None:
+                if scale in enums.SI_SCALE_CODE.values:
+                    item_base.scale = scale
             else:
                 raise ValueError("The 'scale' argument must be one of '{'. ',join(enums.SI_SCALE_CODE.values)}")
 
         # Check if unit is compatible
-        if unit is not None and unit != item_base.item_units \
-                and unit not in item_base.acceptable_units:
+        if unit is not None and unit != item_base.unit and unit not in item_base.acceptable_units:
             logger.warning(f"The supplied unit {unit} for measurement {measurement} "
-                           f"will be ignored, {item_base.item_units} will be used instead."
+                           f"will be ignored, {item_base.unit} will be used instead. "
                            f"Allowed units for this measurement are: "
                            f"{', '.join(item_base.acceptable_units)}")
 
         # Get or create the relevant Report
         if report_specifier_id:
-            report = find_by(self.reports,
-                             'report_name', report_name,
-                             'report_specifier_id', report_specifier_id)
+            report = utils.find_by(self.reports,
+                                   'report_name', report_name,
+                                   'report_specifier_id', report_specifier_id)
         else:
-            report = find_by(self.reports, 'report_name', report_name)
+            report = utils.find_by(self.reports, 'report_name', report_name)
 
         if not report:
-            report_specifier_id = report_specifier_id or generate_id()
+            report_specifier_id = report_specifier_id or utils.generate_id()
             report = objects.Report(created_date_time=datetime.now(),
                                     report_name=report_name,
                                     report_specifier_id=report_specifier_id,
@@ -298,7 +291,7 @@ class OpenADRClient:
 
         # Add the new report description to the report
         target = objects.Target(resource_id=resource_id)
-        r_id = generate_id()
+        r_id = utils.generate_id()
         report_description = objects.ReportDescription(r_id=r_id,
                                                        reading_type=reading_type,
                                                        report_data_source=target,
@@ -335,7 +328,7 @@ class OpenADRClient:
         """
         Request information about the VTN.
         """
-        request_id = generate_id()
+        request_id = utils.generate_id()
         service = 'EiRegisterParty'
         message = self._create_message('oadrQueryRegistration', request_id=request_id)
         response_type, response_payload = await self._perform_request(service, message)
@@ -359,7 +352,7 @@ class OpenADRClient:
         :param str ven_id: The ID for this VEN. If you leave this blank,
                            a VEN_ID will be assigned by the VTN.
         """
-        request_id = generate_id()
+        request_id = utils.generate_id()
         service = 'EiRegisterParty'
         payload = {'ven_name': self.ven_name,
                    'http_pull_model': http_pull_model,
@@ -403,7 +396,7 @@ class OpenADRClient:
         """
         Request the next Event from the VTN, if it has any.
         """
-        payload = {'request_id': generate_id(),
+        payload = {'request_id': utils.generate_id(),
                    'ven_id': self.ven_id,
                    'reply_limit': reply_limit}
         message = self._create_message('oadrRequestEvent', **payload)
@@ -440,7 +433,7 @@ class OpenADRClient:
         Tell the VTN about our reports. The VTN miht respond with an
         oadrCreateReport message that tells us which reports are to be sent.
         """
-        request_id = generate_id()
+        request_id = utils.generate_id()
         payload = {'request_id': request_id,
                    'ven_id': self.ven_id,
                    'reports': reports}
@@ -473,7 +466,7 @@ class OpenADRClient:
         granularity = report_request['report_specifier']['granularity']
 
         # Check if this report actually exists
-        report = find_by(self.reports, 'report_specifier_id', report_specifier_id)
+        report = utils.find_by(self.reports, 'report_specifier_id', report_specifier_id)
         if not report:
             logger.error(f"A non-existant report with report_specifier_id "
                          f"{report_specifier_id} was requested.")
@@ -484,7 +477,7 @@ class OpenADRClient:
         for specifier_payload in report_request['report_specifier']['specifier_payloads']:
             r_id = specifier_payload['r_id']
             # Check if the requested r_id actually exists
-            rd = find_by(report.report_descriptions, 'r_id', r_id)
+            rd = utils.find_by(report.report_descriptions, 'r_id', r_id)
             if not rd:
                 logger.error(f"A non-existant report with r_id {r_id} "
                              f"inside report with report_specifier_id {report_specifier_id} "
@@ -494,17 +487,17 @@ class OpenADRClient:
             # Check if the requested measurement exists and if the correct unit is requested
             if 'measurement' in specifier_payload:
                 measurement = specifier_payload['measurement']
-                if measurement['item_description'] != rd.measurement.item_description:
+                if measurement['description'] != rd.measurement.description:
                     logger.error(f"A non-matching measurement description for report with "
                                  f"report_request_id {report_request_id} and r_id {r_id} was given "
-                                 f"by the VTN. Offered: {rd.measurement.item_description}, "
-                                 f"requested: {measurement['item_description']}")
+                                 f"by the VTN. Offered: {rd.measurement.description}, "
+                                 f"requested: {measurement['description']}")
                     continue
-                if measurement['item_units'] != rd.measurement.item_units:
+                if measurement['unit'] != rd.measurement.unit:
                     logger.error(f"A non-matching measurement unit for report with "
                                  f"report_request_id {report_request_id} and r_id {r_id} was given "
-                                 f"by the VTN. Offered: {rd.measurement.item_units}, "
-                                 f"requested: {measurement['item_units']}")
+                                 f"by the VTN. Offered: {rd.measurement.unit}, "
+                                 f"requested: {measurement['unit']}")
                     continue
 
             if granularity is not None:
@@ -526,7 +519,7 @@ class OpenADRClient:
         reporting_interval = report_back_duration or granularity
         job = self.scheduler.add_job(func=callback,
                                      trigger='cron',
-                                     **cron_config(reporting_interval))
+                                     **utils.cron_config(reporting_interval))
 
         self.report_requests.append({'report_request_id': report_request_id,
                                      'report_specifier_id': report_specifier_id,
@@ -535,16 +528,21 @@ class OpenADRClient:
                                      'granularity': granularity,
                                      'job': job})
 
+    async def create_single_report(self, report_request):
+        """
+        Create a single report in response to a request from the VTN.
+        """
+
     async def update_report(self, report_request_id):
         """
         Call the previously registered report callback and send the result as a message to the VTN.
         """
         logger.debug(f"Running update_report for {report_request_id}")
-        report_request = find_by(self.report_requests, 'report_request_id', report_request_id)
+        report_request = utils.find_by(self.report_requests, 'report_request_id', report_request_id)
         granularity = report_request['granularity']
         report_back_duration = report_request['report_back_duration']
         report_specifier_id = report_request['report_specifier_id']
-        report = find_by(self.reports, 'report_specifier_id', report_specifier_id)
+        report = utils.find_by(self.reports, 'report_specifier_id', report_specifier_id)
         data_collection_mode = report.data_collection_mode
 
         if report_request_id in self.incomplete_reports:
@@ -599,13 +597,23 @@ class OpenADRClient:
             expected_len = len(report_request['r_ids']) * int(report_interval / sampling_interval)
             if len(outgoing_report.intervals) == expected_len:
                 logger.info("The report is now complete with all the values. Will queue for sending.")
-                await self.pending_reports.put(self.incomplete_reports.pop(report_request_id))
+                if self.allow_jitter:
+                    delay = random.uniform(0, min(30, report_interval / 2))
+                    self.loop.create_task(utils.delayed_call(func=self.pending_reports.put(outgoing_report),
+                                                             delay=delay))
+                else:
+                    await self.pending_reports.put(self.incomplete_reports.pop(report_request_id))
             else:
                 logger.debug("The report is not yet complete, will hold until it is.")
                 self.incomplete_reports[report_request_id] = outgoing_report
         else:
             logger.info("Report will be sent now.")
-            await self.pending_reports.put(outgoing_report)
+            if self.allow_jitter:
+                delay = random.uniform(0, min(30, granularity.total_seconds() / 2))
+                self.loop.create_task(utils.delayed_call(func=self.pending_reports.put(outgoing_report),
+                                                         delay=delay))
+            else:
+                await self.pending_reports.put(outgoing_report)
 
     async def cancel_report(self, payload):
         """
@@ -619,7 +627,6 @@ class OpenADRClient:
 
         while True:
             report = await self.pending_reports.get()
-
             service = 'EiReport'
             message = self._create_message('oadrUpdateReport', reports=[report])
 
@@ -646,6 +653,16 @@ class OpenADRClient:
                        "choice. Will opt out of the event for now.")
         return 'optOut'
 
+    async def on_update_event(self, event):
+        """
+        Placeholder for the on_update_event handler.
+        """
+        logger.warning("You should implement your own on_update_event handler. This handler receives "
+                       "an Event dict and should return either 'optIn' or 'optOut' based on your "
+                       "choice. Will re-use the previous opt status for this event_id for now")
+        if event['event_descriptor']['event_id'] in self.events:
+            return self.responded_events['event_id']
+
     ###########################################################################
     #                                                                         #
     #                                  LOW LEVEL                              #
@@ -660,8 +677,8 @@ class OpenADRClient:
             async with self.client_session.post(url, data=message) as req:
                 content = await req.read()
                 if req.status != HTTPStatus.OK:
-                    logger.warning(f"Non-OK status when performing a request to {url} with data "
-                                   f"{message}: {req.status} {content.decode('utf-8')}")
+                    logger.warning(f"Non-OK status {req.status} when performing a request to {url} "
+                                   f"with data {message}: {req.status} {content.decode('utf-8')}")
                     return None, {}
                 logger.debug(content.decode('utf-8'))
         except aiohttp.client_exceptions.ClientConnectorError as err:
@@ -689,18 +706,34 @@ class OpenADRClient:
         except Exception as err:
             logger.error(f"The incoming message could not be parsed or validated: {err}")
             return None, {}
+        if message_payload['response']['response_code'] != 200:
+            logger.warning("We got a non-OK OpenADR response from the server: "
+                           f"{message_payload['response']['response_code']}: "
+                           f"{message_payload['response']['response_description']}")
         return message_type, message_payload
 
     async def _on_event(self, message):
         logger.debug("The VEN received an event")
         events = message['events']
         try:
-            results = [self.on_event(event) for event in message['events']]
-            if asyncio.iscoroutine(results[0]):
-                results = await asyncio.gather(*results, return_exceptions=False)
+            results = []
+            for event in message['events']:
+                event_id = event['event_descriptor']['event_id']
+                event_status = event['event_descriptor']['event_status']
+                if event_id in self.responded_events:
+                    result = self.on_update_event(event)
+                else:
+                    result = self.on_event(event)
+                if asyncio.iscoroutine(result):
+                    result = await result
+                results.append(result)
+                if event_status == 'completed':
+                    self.responded_events.pop(event_id)
+                else:
+                    self.responded_events[event_id] = result
             for i, result in enumerate(results):
                 if result not in ('optIn', 'optOut'):
-                    logger.error("Your on_event handler must return 'optIn' or 'optOut'; "
+                    logger.error("Your on_event or on_update_event handler must return 'optIn' or 'optOut'; "
                                  f"you supplied {result}. Please fix your on_event handler.")
                     results[i] = 'optOut'
         except Exception as err:
@@ -708,7 +741,6 @@ class OpenADRClient:
                          f"The error was {err.__class__.__name__}: {str(err)}")
             results = ['optOut'] * len(events)
 
-        print("Done executing on_event for events")
         if len(events) == 1:
             logger.debug(f"Now responding with {results[0]}")
         else:
@@ -727,7 +759,8 @@ class OpenADRClient:
                     'request_id': message['request_id']}
         message = self._create_message('oadrCreatedEvent',
                                        response=response,
-                                       event_responses=event_responses)
+                                       event_responses=event_responses,
+                                       ven_id=self.ven_id)
         service = 'EiEvent'
         response_type, response_payload = await self._perform_request(service, message)
         logger.info(response_type, response_payload)
@@ -753,6 +786,11 @@ class OpenADRClient:
         elif response_type == 'oadrUpdateReport':
             await self._on_report(response_payload)
 
+        elif response_type == 'oadrCreateReport':
+            if 'report_requests' in response_payload:
+                for report_request in response_payload['report_requests']:
+                    await self.create_report(report_request)
+
         else:
             logger.warning(f"No handler implemented for incoming message "
                            f"of type {response_type}, ignoring.")
