@@ -19,12 +19,13 @@ from aiohttp import web
 from openleadr.service import EventService, PollService, RegistrationService, ReportService, \
                               OptService, VTNService
 from openleadr.messaging import create_message
-from openleadr.utils import certificate_fingerprint, generate_id
 from openleadr import objects
+from openleadr import utils
 from functools import partial
 from datetime import datetime, timedelta, timezone
 import logging
 import ssl
+import re
 logger = logging.getLogger('openleadr')
 
 
@@ -108,7 +109,7 @@ class OpenADRServer:
                 print("")
                 print("*" * 80)
                 print("Your VTN Certificate Fingerprint is "
-                      f"{certificate_fingerprint(cert)}".center(80))
+                      f"{utils.certificate_fingerprint(cert)}".center(80))
                 print("Please deliver this fingerprint to the VENs that connect to you.".center(80))
                 print("You do not need to keep this a secret.".center(80))
                 print("*" * 80)
@@ -149,7 +150,9 @@ class OpenADRServer:
     async def stop(self):
         await self.app_runner.cleanup()
 
-    def add_event(self, ven_id, signal_name, signal_type, intervals, target, callback):
+    def add_event(self, ven_id, signal_name, signal_type, intervals, callback, targets=None,
+                  targets_by_type=None, target=None, market_context="oadr://unknown.context",
+                  notification_period=None, ramp_up_period=None, recovery_period=None):
         """
         Convenience method to add an event with a single signal.
         :param str ven_id: The ven_id to whom this event must be delivered.
@@ -157,6 +160,15 @@ class OpenADRServer:
         :param str signal_type: The OpenADR type of the signal; one of openleadr.objects.SIGNAL_TYPE
         :param str intervals: A list of intervals with a dtstart, duration and payload member.
         :param str callback: A callback function for when your event has been accepted (optIn) or refused (optOut).
+        :param list targets: A list of Targets that this Event applies to.
+        :param target: A single target for this event.
+        :param dict targets_by_type: A dict of targets, grouped by type.
+        :param str market_context: A URI for the DR program that this event belongs to.
+        :param timedelta notification_period: The Notification period for the Event's Active Period.
+        :param timedelta ramp_up_period: The Ramp Up period for the Event's Active Period.
+        :param timedelta recovery_period: The Recovery period for the Event's Active Period.
+
+        If you don't provide a target using any of the three arguments, the target will be set to the given ven_id.
         """
         if self.services['event_service'].polling_method == 'external':
             logger.error("You cannot use the add_event method after you assign your own on_poll "
@@ -165,26 +177,53 @@ class OpenADRServer:
                          "message queuing system, you should not assign an on_poll handler. "
                          "Your Event will NOT be added.")
             return
-        event_id = generate_id()
-        if not isinstance(target, list):
-            target = [target]
+        if not re.match(r"^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?", market_context):
+            raise ValueError("The Market Context must be a valid URI.")
+        event_id = utils.generate_id()
+
+        # Figure out the target for this Event
+        if target is None and targets is None and targets_by_type is None:
+            targets = [{'ven_id': ven_id}]
+        elif target is not None:
+            targets = [target]
+        elif targets_by_type is not None:
+            targets = utils.ungroup_targets_by_type(targets_by_type)
+        if not isinstance(targets, list):
+            targets = [targets]
+
         event_descriptor = objects.EventDescriptor(event_id=event_id,
                                                    modification_number=0,
-                                                   market_context="None",
-                                                   event_status="near",
+                                                   market_context=market_context,
+                                                   event_status="far",
                                                    created_date_time=datetime.now(timezone.utc))
         event_signal = objects.EventSignal(intervals=intervals,
                                            signal_name=signal_name,
                                            signal_type=signal_type,
-                                           signal_id=generate_id(),
-                                           targets=target)
-        event = objects.Event(event_descriptor=event_descriptor,
+                                           signal_id=utils.generate_id(),
+                                           targets=targets)
+        # Make sure the intervals carry timezone-aware timestamps
+        for interval in intervals:
+            if utils.getmember(interval, 'dtstart').tzinfo is None:
+                utils.setmember(interval, 'dtstart',
+                                utils.getmember(interval, 'dtstart').astimezone(timezone.utc))
+                logger.warning("You supplied a naive datetime object to your interval's dtstart. "
+                               "This will be interpreted as a timestamp in your local timezone "
+                               "and then converted to UTC before sending. Please supply timezone-"
+                               "aware timestamps like datetime.datetime.new(timezone.utc) or "
+                               "datetime.datetime(..., tzinfo=datetime.timezone.utc)")
+        active_period = utils.get_active_period_from_intervals(intervals, False)
+        active_period.ramp_up_period = ramp_up_period
+        active_period.notification_period = notification_period
+        active_period.recovery_period = recovery_period
+        event = objects.Event(active_period=active_period,
+                              event_descriptor=event_descriptor,
                               event_signals=[event_signal],
-                              targets=target)
+                              targets=targets)
         if ven_id not in self.message_queues:
             self.message_queues[ven_id] = asyncio.Queue()
         self.message_queues[ven_id].put_nowait(event)
-        self.services['event_service'].pending_events[event_id] = callback
+        self.services['event_service'].pending_events[event_id] = (event, callback)
+        return event_id
 
     def add_raw_event(self, ven_id, event):
         """
