@@ -15,9 +15,11 @@
 # limitations under the License.
 
 from . import service, handler, VTNService
-from asyncio import iscoroutine
-from .. import objects
+import asyncio
+from openleadr import objects, utils, enums
 import logging
+from datetime import datetime, timezone
+from functools import partial
 logger = logging.getLogger('openleadr')
 
 
@@ -29,6 +31,7 @@ class EventService(VTNService):
         self.polling_method = polling_method
         self.message_queues = message_queues
         self.pending_events = {}        # Holds the event callbacks
+        self.running_events = {}        # Holds the event callbacks for accepted events
 
     @handler('oadrRequestEvent')
     async def request_event(self, payload):
@@ -36,7 +39,7 @@ class EventService(VTNService):
         The VEN requests us to send any events we have.
         """
         result = self.on_request_event(payload['ven_id'])
-        if iscoroutine(result):
+        if asyncio.iscoroutine(result):
             result = await result
         if result is None:
             return 'oadrDistributeEvent', {'events': []}
@@ -64,19 +67,44 @@ class EventService(VTNService):
         """
         The VEN informs us that they created an EiEvent.
         """
+        loop = asyncio.get_event_loop()
+        ven_id = payload['ven_id']
         if self.polling_method == 'internal':
             for event_response in payload['event_responses']:
+                event_id = event_response['event_id']
+                opt_type = event_response['opt_type']
                 if event_response['event_id'] in self.pending_events:
-                    ven_id = payload['ven_id']
-                    event_id = event_response['event_id']
-                    opt_type = event_response['opt_type']
-                    callback = self.pending_events.pop(event_id)
+                    event, callback = self.pending_events.pop(event_id)
                     result = callback(ven_id=ven_id, event_id=event_id, opt_type=opt_type)
-                    if iscoroutine(result):
+                    if asyncio.iscoroutine(result):
+                        result = await result
+                    if opt_type == 'optIn':
+                        self.running_events[event_id] = (event, callback)
+                        now = datetime.now(timezone.utc)
+                        active_period = event.active_period
+                        # Schedule status update to 'near' if applicable
+                        if active_period.ramp_up_period is not None and event.event_descriptor.event_status == 'far':
+                            ramp_up_start_delay = (active_period.dtstart - active_period.ramp_up_period) - now
+                            update_coro = partial(self._update_event_status, ven_id, event, 'near')
+                            loop.create_task(utils.delayed_call(func=update_coro, delay=ramp_up_start_delay))
+                        # Schedule status update to 'active'
+                        if event.event_descriptor.event_status in ('near', 'far'):
+                            active_start_delay = active_period.dtstart - now
+                            update_coro = partial(self._update_event_status, ven_id, event, 'active')
+                            loop.create_task(utils.delayed_call(func=update_coro, delay=active_start_delay))
+                        # Schedule status update to 'completed'
+                        if event.event_descriptor.event_status in ('near', 'far', 'active'):
+                            active_end_delay = active_period.dtstart + active_period.duration - now
+                            update_coro = partial(self._update_event_status, ven_id, event, 'completed')
+                            loop.create_task(utils.delayed_call(func=update_coro, delay=active_end_delay))
+                elif event_response['event_id'] in self.running_events:
+                    event, callback = self.running_events.pop(event_id)
+                    result = callback(ven_id=ven_id, event_id=event_id, opt_type=opt_type)
+                    if asyncio.iscoroutine(result):
                         result = await result
         else:
             result = self.on_created_event(ven_id=ven_id, event_id=event_id, opt_type=opt_type)
-            if iscoroutine(result):
+            if asyncio.iscoroutine(result):
                 result = await(result)
         return 'oadrResponse', {}
 
@@ -89,3 +117,12 @@ class EventService(VTNService):
                        "handler will receive a ven_id, event_id and opt_status. "
                        "You don't need to return anything from this handler.")
         return None
+
+    def _update_event_status(self, ven_id, event, event_status):
+        """
+        Update the event to the given Status.
+        """
+        event.event_descriptor.event_status = event_status
+        if event_status == enums.EVENT_STATUS.CANCELLED:
+            event.event_descriptor.modification_number += 1
+        self.message_queues[ven_id].put_nowait(event)
