@@ -136,11 +136,12 @@ class VTNService:
             response_payload['request_id'] = generate_id()
 
         else:
-            response_type, response_payload = self.error_response(message_type,
+            response_type, response_payload = self.error_response('oadrResponse',
                                                                   STATUS_CODES.COMPLIANCE_ERROR,
                                                                   "A message of type "
                                                                   f"{message_type} should not be "
                                                                   "sent to this endpoint")
+        logger.info(f"Responding to {message_type} with a {response_type} message: {response_payload}.")
         return response_type, response_payload
 
     def error_response(self, message_type, error_code, error_description):
@@ -151,5 +152,5 @@ class VTNService:
         else:
             response_type = 'oadrResponse'
         response_payload = {'response': {'response_code': error_code,
-                                         'response_description': 'Certificate fingerprint mismatch'}}
+                                         'response_description': error_description}}
         return response_type, response_payload
