@@ -17,11 +17,16 @@
 from datetime import datetime, timedelta, timezone
 from dataclasses import is_dataclass, asdict
 from collections import OrderedDict
+from openleadr import enums
+import asyncio
 import itertools
 import re
 import ssl
 import hashlib
 import uuid
+import logging
+
+logger = logging.getLogger('openleadr')
 
 DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
 DATETIME_FORMAT_NO_MICROSECONDS = "%Y-%m-%dT%H:%M:%SZ"
@@ -78,6 +83,9 @@ def normalize_dict(ordered_dict):
             key = key[4:]
         elif key.startswith('ei'):
             key = key[2:]
+        # Don't normalize the measurement descriptions
+        if key in enums._MEASUREMENT_NAMESPACES:
+            return key
         key = re.sub(r'([a-z])([A-Z])', r'\1_\2', key)
         if '-' in key:
             key = key.replace('-', '_')
@@ -123,7 +131,7 @@ def normalize_dict(ordered_dict):
             key = key[5:]
 
         # Group all targets as a list of dicts under the key "target"
-        if key in ("target",):
+        if key == 'target':
             targets = d.pop(key)
             new_targets = []
             if targets:
@@ -135,6 +143,10 @@ def normalize_dict(ordered_dict):
             d[key + "s"] = new_targets
             key = key + "s"
 
+            # Also add a targets_by_type element to this dict
+            # to access the targets in a more convenient way.
+            d['targets_by_type'] = group_targets_by_type(new_targets)
+
         # Group all reports as a list of dicts under the key "pending_reports"
         if key == "pending_reports":
             if isinstance(d[key], dict) and 'report_request_id' in d[key] \
@@ -175,41 +187,19 @@ def normalize_dict(ordered_dict):
                 descriptions = [descriptions]
             for description in descriptions:
                 # We want to make the identification of the measurement universal
-                if 'voltage' in description:
-                    item_name, item = 'voltage', description.pop('voltage')
-                elif 'power_real' in description:
-                    item_name, item = 'powerReal', description.pop('power_real')
-                elif 'power_apparent' in description:
-                    item_name, item = 'powerApparent', description.pop('power_apparent')
-                elif 'power_reactive' in description:
-                    item_name, item = 'powerReactive', description.pop('power_reactive')
-                elif 'energy_real' in description:
-                    item_name, item = 'energyReal', description.pop('energy_real')
-                elif 'energy_apparent' in description:
-                    item_name, item = 'energyApparent', description.pop('energy_apparent')
-                elif 'energy_reactive' in description:
-                    item_name, item = 'energyReactive', description.pop('energy_reactive')
-                elif 'frequency' in description:
-                    item_name, item = 'frequency', description.pop('frequency')
-                elif 'pulse_count' in description:
-                    item_name, item = 'pulseCount', description.pop('pulse_count')
-                elif 'temperature' in description:
-                    item_name, item = 'temperature', description.pop('temperature')
-                elif 'therm' in description:
-                    item_name, item = 'therm', description.pop('therm')
-                elif 'currency' in description:
-                    item_name, item = 'currency', description.pop('currency')
-                elif 'currency_per_kw' in description:
-                    item_name, item = 'currencyPerKW', description.pop('currency_per_kw')
-                elif 'currency_per_kwh' in description:
-                    item_name, item = 'currencyPerKWh', description.pop('currency_per_kwh')
-                elif 'currency_per_therm' in description:
-                    item_name, item = 'currencyPerTherm', description.pop('currency_per_therm')
-                elif 'custom_unit' in description:
-                    item_name, item = 'customUnit', description.pop('custom_unit')
+                for measurement in enums._MEASUREMENT_NAMESPACES:
+                    if measurement in description:
+                        name, item = measurement, description.pop(measurement)
+                        break
                 else:
                     break
-                description['measurement'] = {'item_name': item_name,
+                item['description'] = item.pop('item_description', None)
+                item['unit'] = item.pop('item_units', None)
+                if 'si_scale_code' in item:
+                    item['scale'] = item.pop('si_scale_code')
+                if 'pulse_factor' in item:
+                    item['pulse_factor'] = item.pop('pulse_factor')
+                description['measurement'] = {'name': name,
                                               **item}
             d[key + 's'] = descriptions
 
@@ -288,7 +278,7 @@ def parse_datetime(value):
         year, month, day, hour, minute, second, micro = (int(value) for value in matches.groups())
         return datetime(year, month, day, hour, minute, second, micro, tzinfo=timezone.utc)
     else:
-        print(f"{value} did not match format")
+        logger.warning(f"parse_datetime: {value} did not match format")
         return value
 
 
@@ -482,7 +472,7 @@ def group_by(list_, key, pop_key=False):
     return grouped
 
 
-def cron_config(interval):
+def cron_config(interval, randomize_seconds=False):
     """
     Returns a dict with cron settings for the given interval
     """
@@ -492,13 +482,21 @@ def cron_config(interval):
         hour = "*"
     elif interval < timedelta(hours=1):
         second = "0"
-        minute = f"*/{int(interval.total_seconds/60)}"
+        minute = f"*/{int(interval.total_seconds()/60)}"
         hour = "*"
-    elif interval < timedelta(days=1):
+    elif interval < timedelta(hours=24):
         second = "0"
         minute = "0"
-        hour = f"*/{int(interval.total_seconds/3600)}"
-    return {"second": second, "minute": minute, "hour": hour}
+        hour = f"*/{int(interval.total_seconds()/3600)}"
+    else:
+        second = "0"
+        minute = "0"
+        hour = "0"
+    cron_config = {"second": second, "minute": minute, "hour": hour}
+    if randomize_seconds:
+        jitter = min(int(interval.total_seconds() / 10), 300)
+        cron_config['jitter'] = jitter
+    return cron_config
 
 
 def get_cert_fingerprint_from_request(request):
@@ -514,3 +512,155 @@ def get_certificate_common_name(request):
     if cert:
         subject = dict(x[0] for x in cert['subject'])
         return subject.get('commonName')
+
+
+def group_targets_by_type(list_of_targets):
+    targets_by_type = {}
+    for target in list_of_targets:
+        for key, value in target.items():
+            if value is None:
+                continue
+            if key not in targets_by_type:
+                targets_by_type[key] = []
+            targets_by_type[key].append(value)
+    return targets_by_type
+
+
+def ungroup_targets_by_type(targets_by_type):
+    ungrouped_targets = []
+    for target_type, targets in targets_by_type.items():
+        if isinstance(targets, list):
+            for target in targets:
+                ungrouped_targets.append({target_type: target})
+        elif isinstance(targets, str):
+            ungrouped_targets.append({target_type: targets})
+    return ungrouped_targets
+
+
+def validate_report_measurement_dict(measurement):
+    from openleadr.enums import _ACCEPTABLE_UNITS, _MEASUREMENT_DESCRIPTIONS
+
+    if 'name' not in measurement \
+            or 'description' not in measurement \
+            or 'unit' not in measurement:
+        raise ValueError("The measurement dict must contain the following keys: "
+                         "'name', 'description', 'unit'. Please correct this.")
+
+    name = measurement['name']
+    description = measurement['description']
+    unit = measurement['unit']
+
+    # Validate the item name and description match
+    if name in _MEASUREMENT_DESCRIPTIONS:
+        required_description = _MEASUREMENT_DESCRIPTIONS[name]
+        if description != required_description:
+            if description.lower() == required_description.lower():
+                logger.warning(f"The description for the measurement with name {name} "
+                               f"was not in the correct case; you provided {description} but "
+                               f"it should be {required_description}. "
+                               "This was automatically corrected.")
+                measurement['description'] = required_description
+            else:
+                raise ValueError(f"The measurement's description {description} "
+                                 f"did not match the expected description for this type "
+                                 f" ({required_description}). Please correct this, or use "
+                                 "'customUnit' as the name.")
+        if unit not in _ACCEPTABLE_UNITS[name]:
+            raise ValueError(f"The unit {unit} is not acceptable for measurement {name}. Allowed "
+                             f"units are {_ACCEPTABLE_UNITS[name]}.")
+    else:
+        if name != 'customUnit':
+            logger.warning(f"You provided a measurement with an unknown name {name}. "
+                           "This was corrected to 'customUnit'. Please correct this in your "
+                           "report definition.")
+            measurement['report_description']['name'] = 'customUnit'
+
+    if 'power' in name:
+        if 'power_attributes' in measurement:
+            power_attributes = measurement['power_attributes']
+            if 'voltage' not in power_attributes \
+                    or 'ac' not in power_attributes \
+                    or 'hertz' not in power_attributes:
+                raise ValueError("The power_attributes of the measurement must contain the "
+                                 "following keys: 'voltage' (int), 'ac' (bool), 'hertz' (int).")
+        else:
+            raise ValueError("A 'power' related measurement must contain a "
+                             "'power_attributes' section that contains the following "
+                             "keys: 'voltage' (int), 'ac' (boolean), 'hertz' (int)")
+
+
+def get_active_period_from_intervals(intervals, as_dict=True):
+    if is_dataclass(intervals[0]):
+        intervals = [asdict(i) for i in intervals]
+    period_start = min([i['dtstart'] for i in intervals])
+    period_duration = max([i['dtstart'] + i['duration'] - period_start for i in intervals])
+    if as_dict:
+        return {'dtstart': period_start,
+                'duration': period_duration}
+    else:
+        from openleadr.objects import ActivePeriod
+        return ActivePeriod(dtstart=period_start, duration=period_duration)
+
+
+def determine_event_status(active_period):
+    if is_dataclass(active_period):
+        active_period = asdict(active_period)
+    now = datetime.now(timezone.utc)
+    if active_period['dtstart'].tzinfo is None:
+        active_period['dtstart'] = active_period['dtstart'].astimezone(timezone.utc)
+    active_period_start = active_period['dtstart']
+    active_period_end = active_period['dtstart'] + active_period['duration']
+    if now >= active_period_end:
+        return 'completed'
+    if now >= active_period_start:
+        return 'active'
+    if active_period.get('ramp_up_duration') is not None:
+        ramp_up_start = active_period_start - active_period['ramp_up_duration']
+        if now >= ramp_up_start:
+            return 'near'
+    return 'far'
+
+
+async def delayed_call(func, delay):
+    if isinstance(delay, timedelta):
+        delay = delay.total_seconds()
+    await asyncio.sleep(delay)
+    if asyncio.iscoroutinefunction(func):
+        await func()
+    elif asyncio.iscoroutine(func):
+        await func
+    else:
+        func()
+
+
+def hasmember(obj, member):
+    """
+    Check if a dict or dataclass has the given member
+    """
+    if is_dataclass(obj):
+        if hasattr(obj, member):
+            return True
+    else:
+        if member in obj:
+            return True
+    return False
+
+
+def getmember(obj, member):
+    """
+    Get a member from a dict or dataclass
+    """
+    if is_dataclass(obj):
+        return getattr(obj, member)
+    else:
+        return obj[member]
+
+
+def setmember(obj, member, value):
+    """
+    Set a member of a dict of dataclass
+    """
+    if is_dataclass(obj):
+        setattr(obj, member, value)
+    else:
+        obj[member] = value
